
/* Patch distance templates for similarity transform (rotation+scale). */

#ifndef _simpatch_h
#define _simpatch_h

#include "simnn.h"

template<int TPATCH_W, int DO_BRANCH>
int sim_fast_patch_dist(int *adata, BITMAP *b, XFORM bpos, int maxval) {
  int ans = 0;
  int bx_row = bpos.x0, by_row = bpos.y0;
  for (int dy = 0; dy < TPATCH_W; dy++) {
    int bx = bx_row, by = by_row;
    for (int dx = 0; dx < TPATCH_W; dx++) {
      unsigned int c1 = adata[dx];
      int r2, g2, b2;
      getpixel_bilin(b, bx, by, r2, g2, b2);
      int dr = (c1&255)-r2;
      int dg = ((c1>>8)&255)-g2;
      int db = (c1>>16)-b2;
      ans += dr*dr+dg*dg+db*db;
      if (DO_BRANCH && ans > maxval) { return ans; }
      bx += bpos.dxdu;
      by += bpos.dydu;
    }
    adata += TPATCH_W;
    bx_row += bpos.dxdv;
    by_row += bpos.dydv;
  }
  return ans;
}

template<int PATCH_W>
void sim_attempt_n(int &err, int &xbest, int &ybest, int &sbest, int &tbest, int *adata, BITMAP *b, XFORM bpos, int bx, int by, int bs, int bt, Params *p) {
  //int h = PATCH_W/2;
  if ((bx != xbest || by != ybest || bs != sbest || bt != tbest) &&
      (unsigned) (bx) < (unsigned) (b->w-PATCH_W+1) &&
      (unsigned) (by) < (unsigned) (b->h-PATCH_W+1)) {
    //XFORM bpos = get_xform(p, bx, by, bs, bt);
    int current = sim_fast_patch_dist<PATCH_W, 1>(adata, b, bpos, err);
    if (current < err) {
      err = current;
      xbest = bx;
      ybest = by;
      sbest = bs;
      tbest = bt;
    }
  }
}

template<int TPATCH_W, int DO_BRANCH>
int sim_patch_dist_ab(Params *p, BITMAP *a, int ax, int ay, BITMAP *b, int bx, int by, int bs, int bt, int maxval) {
  int adata[TPATCH_W*TPATCH_W];
  int *ptr = adata;
  for (int dy = 0; dy < TPATCH_W; dy++) {
    int *row = ((int *) a->line[ay+dy])+ax;
    for (int dx = 0; dx < TPATCH_W; dx++) {
      *ptr++ = *row++;
    }
  }
  XFORM bpos = get_xform(p, bx, by, bs, bt);
  return sim_fast_patch_dist<TPATCH_W, DO_BRANCH>(adata, b, bpos, maxval);
}

#endif
