
/* Patch distance templates for vector input. */

#ifndef _vecpatch_h
#define _vecpatch_h

#include "vecnn.h"

template<class T, class ACCUM, int IS_WINDOW>
ACCUM vec_fast_patch_dist(T *apatch, T *bpatch, ACCUM maxval, Params *p) {
  ACCUM ans = 0;
  for (int i = 0; i < p->vec_len; i++) {
    T c1 = apatch[i];
    T c2 = bpatch[i];
    ACCUM d = ((ACCUM) apatch[i])-((ACCUM) bpatch[i]);
    ans += d*d;
    if (ans > maxval) { return ans; }
  }
  return ans;
}

template<class T, class ACCUM, int IS_MASK, int IS_WINDOW>
void vec_attempt_n(ACCUM &err, int &xbest, int &ybest, T *apatch, VECBITMAP<T> *b, int bx, int by, BITMAP *bmask, RegionMasks *region_masks, int src_mask, Params *p) {
  if ((bx != xbest || by != ybest) &&
      (unsigned) bx < (unsigned) (b->w) &&
      (unsigned) by < (unsigned) (b->h)) {
    if (IS_MASK && region_masks && src_mask != ((int *) region_masks->bmp->line[by])[bx]) { return; }
    if (IS_MASK && bmask && ((int *) bmask->line[by])[bx]) { return; }
    ACCUM current = vec_fast_patch_dist<T, ACCUM, IS_WINDOW>(apatch, b->get(bx, by), err, p);
    if (current < err) {
      err = current;
      xbest = bx;
      ybest = by;
    }
  }
}

template<class T, class ACCUM, int IS_WINDOW>
ACCUM vec_fast_patch_nobranch(T *apatch, T *bpatch, Params *p) {
  ACCUM ans = 0;
  for (int i = 0; i < p->vec_len; i++) {
    T c1 = apatch[i];
    T c2 = bpatch[i];
    ACCUM d = ((ACCUM) apatch[i])-((ACCUM) bpatch[i]);
    ans += d*d;
  }
  return ans;
}

template<class T, class ACCUM, int IS_WINDOW, int HAS_MASKS>
ACCUM vec_patch_dist_ab(Params *p, VECBITMAP<T> *a, int ax, int ay, VECBITMAP<T> *b, int bx, int by, ACCUM maxval, RegionMasks *region_masks) {
  if (HAS_MASKS && region_masks && ((int *) region_masks->bmp->line[ay])[ax] != ((int *) region_masks->bmp->line[by])[bx]) { return get_maxval<ACCUM>(); }
  return vec_fast_patch_nobranch<T, ACCUM, IS_WINDOW>(a->get(ax, ay), b->get(bx, by), p);
}

#endif
